# BlenderBIM Add-on - OpenBIM Blender Add-on
# Copyright (C) 2020, 2021, 2022 Dion Moult <dion@thinkmoult.com>
#
# This file is part of BlenderBIM Add-on.
#
# BlenderBIM Add-on is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BlenderBIM Add-on is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BlenderBIM Add-on.  If not, see <http://www.gnu.org/licenses/>.

from xerparser.reader import Reader
import ifcopenshell
import ifcopenshell.api
import ifcopenshell.util.date
from datetime import datetime, timedelta, date
from .common import ScheduleIfcGenerator


class P6XER2Ifc:
    status_map = {"TK_NotStart": "Not Start", "TK_Complete": "Completed", "TK_Active": "Progress"}

    TASK_TYPE_MAP = {
        "TT_Task": "Task",
        "TT_Rsrc": "Resource Dependent Task",
        "TT_LOE": "Level of Effort",
        "TT_Mile": "Start Milestone",
        "TT_FinMile": "Finish Milestone",
        "TT_WBS": "WBS Summary",
    }

    RELATIONSHIP_TYPE_MAPPING = {
        "PR_SS": "START_START",
        "PR_FS": "FINISH_START",
        "PR_SF": "START_FINISH",
        "PR_FF": "FINISH_FINISH",
    }

    resource_type_map = {
        "RT_Labor": "LABOR",
        "RT_Mat": "MATERIAL",
        # https://www.primaverascheduling.com/category/primavera-resources/
        "RT_Nonlabor": "EQUIPMENT",
        "RT_Equip": "EQUIPMENT",
    }

    def __init__(self):
        self.xer = None
        self.file = None
        self.model = None
        self.work_plan = None
        self.project = {}
        self.calendars = {}
        self.wbs = {}
        self.root_activites = []
        self.activities = {}
        self.relationships = {}
        self.resources = {}
        self.output = None

        self.day_map2 = {
            "1": "Monday",
            "2": "Tuesday",
            "3": "Wednesday",
            "4": "Thursday",
            "5": "Friday",
            "6": "Saturday",
            "7": "Sunday",
        }

    def execute(self):
        self.parse_xer()
        settings = {
            "work_plan": self.work_plan,
            "project": self.project,
            "calendars": self.calendars,
            "wbs": self.wbs,
            "root_activities": self.root_activites,
            "activities": self.activities,
            "relationships": self.relationships,
            "resources": self.resources,
        }
        ifcCreator = ScheduleIfcGenerator(self.file, self.output, settings)
        ifcCreator.create_ifc()

    def parse_xer(self):
        self.model = Reader(self.xer)
        # for project in self.model.projects:
        #     if not self.project.get("Name"):
        self.project["Name"] = self.model.projects._projects[0].proj_short_name
        self.parse_calendar_xer()
        self.parse_wbs_xer()
        self.parse_activity_xer()
        self.parse_relationship_xer()
        self.parse_resource_xer()
        # work_schedule = self.create_work_schedule()
        # self.create_tasks(work_schedule)

    def parse_wbs_xer(self):
        for wbs in self.model.wbss:
            self.wbs[wbs.wbs_id] = {
                "Name": wbs.wbs_name,
                "Code": wbs.wbs_short_name,
                "ParentObjectId": wbs.parent_wbs_id,
                "ifc": None,
                "rel": None,
                "activities": [],
            }
        # print(self.wbs)

    def parse_calendar_xer(self):
        standard_work_week = []
        exceptions = {}
        for cal in self.model.calendars:
            standard_work_week = cal.working_hours
            except_lst = cal.exceptions
            for exception in except_lst:
                month = exceptions.setdefault(exception.year, {}).setdefault(exception.month, {})
                month.setdefault("FullDay", [])
                month.setdefault("WorkTime", [])
                exceptions[exception.year][exception.month]["FullDay"].append(exception.day)

            self.calendars[cal.clndr_id] = {
                "Name": cal.clndr_name,
                "Type": cal.clndr_type,
                "HoursPerDay": cal.day_hr_cnt,
                "StandardWorkWeek": standard_work_week,
                "HolidayOrExceptions": exceptions,
            }

    def parse_activity_xer(self):
        for activity in self.model.activities:
            activity_type = activity.task_type
            if activity_type == "TT_LOE":
                continue
            wbs_id = activity.wbs_id
            if wbs_id:
                self.wbs[wbs_id]["activities"].append(activity.task_id)
            else:
                self.root_activites.append(activity.task_id)
            self.activities[activity.task_id] = {
                "Name": activity.task_name,
                "Identification": activity.task_code,
                "StartDate": activity.start_date,
                "FinishDate": activity.end_date,
                "PlannedDuration": activity.target_drtn_hr_cnt,
                "Status": self.status_map[activity.status_code],
                "CalendarObjectId": activity.clndr_id,
                "ifc": None,
            }

    def parse_relationship_xer(self):
        for rel in self.model.relations:
            predecessor = rel.pred_task_id
            successor = rel.task_id
            if predecessor not in self.activities or successor not in self.activities:
                continue
            self.relationships[rel.task_pred_id] = {
                "PredecessorActivity": predecessor,
                "SuccessorActivity": successor,
                "Type": self.RELATIONSHIP_TYPE_MAPPING[rel.pred_type],
                "Lag": rel.lag_hr_cnt,
            }

    def parse_resource_xer(self):
        for rsrc in self.model.resources:
            self.resources[rsrc.rsrc_id] = {
                "Name": rsrc.rsrc_name,
                "Code": rsrc.rsrc_short_name,
                "ParentObjectId": rsrc.parent_rsrc_id,
                "Type": self.resource_type_map[rsrc.rsrc_type],
                "ifc": None,
                "rel": None,
            }


# TODO: add support for resources
# TODO: consider showing progress bar for better user experience
# TODO: support multiple projects in a single file
# TODO: prompt user to select activities and/or wbs nodes to import instead of the full project



# Ifc4D - IFC scheduling utility
# Copyright (C) 2021-2022 Dion Moult <dion@thinkmoult.com>, Yassine Oualid <yassine@sigmadimensions.com>
#
# This file is part of Ifc4D.
#
# Ifc4D is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ifc4D is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Ifc4D.  If not, see <http://www.gnu.org/licenses/>.


import csv
import ifcopenshell
import ifcopenshell.api
import locale
import re


class Csv2Ifc:
    def __init__(self):
        self.csv = None
        self.file = None
        self.tasks = []
        self.work_schedule = None
        self.work_plan = None
        self.is_schedule_of_rates = False
        self.units = {}

    def execute(self):
        self.parse_csv()
        self.create_ifc()

    def parse_csv(self):
        self.parents = {}
        self.headers = {}
        locale.setlocale(locale.LC_ALL, "")  # set the system locale
        with open(self.csv, "r", encoding="ISO-8859-1") as csv_file:
            reader = csv.reader(csv_file)
            for row in reader:
                if not row[0]:
                    continue
                if row[0] == "Hierarchy":
                    for i, col in enumerate(row):
                        if not col:
                            continue
                        self.headers[col] = i
                    continue
                task = self.get_row_task(row)
                hierarchy_key = int(row[0])
                if hierarchy_key == 1:
                    self.tasks.append(task)
                else:
                    self.parents[hierarchy_key - 1]["children"].append(task)
                self.parents[hierarchy_key] = task

    def parse_task_rel(self, task_relationships_str):
        rels = []
        if task_relationships_str:
            for rel_str in task_relationships_str.split(";"):
                rel_parts = re.split(r"(\d+)", rel_str)
                task_1 = rel_parts[1]
                rel_type = rel_parts[2]
                task_2 = rel_parts[3]
                rels.append(
                    {
                        "rel_type": rel_type,
                        "task_1": task_1,
                        "task_2": task_2,
                    }
                )
        return rels

    def get_row_task(self, row):
        hours_per_day = 8
        hierarchy = row[self.headers["Hierarchy"]]
        identification = row[self.headers["Identification"]]
        task_name = row[self.headers["Name"]]

        task_relationships = self.parse_task_rel(row[self.headers["Relationships"]])

        scheduled_start_date = ifcopenshell.util.date.string_to_date(row[self.headers["ScheduleStart"]]) if row[self.headers["ScheduleStart"]] else None
        scheduled_finish_date = ifcopenshell.util.date.string_to_date(row[self.headers["ScheduleFinish"]]) if row[self.headers["ScheduleFinish"]] else None
        scheduled_duration = ifcopenshell.util.date.string_to_duration(row[self.headers["ScheduleDuration"]]) if row[self.headers["ScheduleDuration"]] else None
        actual_start_date = ifcopenshell.util.date.string_to_date(row[self.headers["ActualStart"]]) if row[self.headers["ActualStart"]] else None
        actual_finish_date = ifcopenshell.util.date.string_to_date(row[self.headers["ActualFinish"]]) if row[self.headers["ActualFinish"]] else None
        actual_duration = ifcopenshell.util.date.string_to_duration(row[self.headers["ActualDuration"]]) if row[self.headers["ActualDuration"]] else None

        return {
            "Hierarchy": hierarchy,
            "Identification": identification,
            "Name": task_name,
            "Relationships": task_relationships,
            "ScheduleStart": scheduled_start_date,
            "ScheduleFinish": scheduled_finish_date,
            "ScheduleDuration": scheduled_duration,
            "ActualStart": actual_start_date,
            "ActualFinish": actual_finish_date,
            "ActualDuration": actual_duration,
            "children": [],
        }

    def create_ifc(self):
        if not self.file:
            self.create_boilerplate_ifc()
        if not self.work_plan:
            self.work_plan = ifcopenshell.api.run("sequence.add_work_plan", self.file)
        self.work_schedule = self.create_work_schedule()
        self.create_tasks(self.tasks)
        self.create_rel_sequences(self.tasks)

    def create_boilerplate_ifc(self):
        self.file = ifcopenshell.file(schema="IFC4")
        ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcProject")
        self.work_plan = self.file.create_entity("IfcWorkPlan")

    def create_tasks(self, tasks, parent=None):
        for task in tasks:
            self.create_task(task, parent)

    def create_rel_sequences(self, tasks=None):
        self.sequence_type_map = {
            "FF": "FINISH_FINISH",
            "FS": "FINISH_START",
            "SF": "START_FINISH",
            "SS": "START_START",
        }
        for task in tasks:
            self.create_rel_sequence(task)

    def create_work_schedule(self):
        return ifcopenshell.api.run("sequence.add_work_schedule", self.file, name="import", work_plan=self.work_plan)

    def create_task(self, task, parent_task=None):
        if parent_task is None:
            task["ifc"] = ifcopenshell.api.run("sequence.add_task", self.file, work_schedule=self.work_schedule)
        else:
            task["ifc"] = ifcopenshell.api.run("sequence.add_task", self.file, parent_task=parent_task)

        ifcopenshell.api.run(
            "sequence.edit_task",
            self.file,
            task=task["ifc"],
            attributes={
                "Name": task["Name"],
                "Identification": task["Identification"],
                # "IsMilestone": task["ScheduleStart"] == task["ScheduleFinish"],
            },
        )
        task_time = ifcopenshell.api.run("sequence.add_task_time", self.file, task=task["ifc"])
        ifcopenshell.api.run(
            "sequence.edit_task_time",
            self.file,
            task_time=task_time,
            attributes={
                "ScheduleStart": task["ScheduleStart"],
                "ScheduleFinish": task["ScheduleFinish"],
                "DurationType": "WORKTIME" if task["ScheduleDuration"] else None,
                "ScheduleDuration": task["ScheduleDuration"],
                "ActualStart": task["ActualStart"],
                "ActualFinish": task["ActualFinish"],
                "ActualDuration": task["ActualDuration"],
            },
        )
        self.create_tasks(task["children"], task["ifc"])

    def get_entity_by_identification(self, tasks, task_identification):
        for task in tasks or []:
            if task["Identification"].replace(".", "").replace(" ", "") == task_identification.replace(" ", ""):
                return task["ifc"]
            else:
                entity = self.get_entity_by_identification(task["children"], task_identification)
                if entity:
                    return entity
        return None

    def create_rel_sequence(self, task=None):
        for rel in task["Relationships"] or []:
            task_1 = self.get_entity_by_identification(self.tasks, rel["task_1"])
            task_2 = self.get_entity_by_identification(self.tasks, rel["task_2"])
            rel_type = self.sequence_type_map[rel["rel_type"]]
            rel_sequence = ifcopenshell.api.run(
                "sequence.assign_sequence",
                self.file,
                related_process=task_2,
                relating_process=task_1,
            )
            if rel_type:
                ifcopenshell.api.run(
                    "sequence.edit_sequence",
                    self.file,
                    rel_sequence=rel_sequence,
                    attributes={"SequenceType": rel_type},
                )
        self.create_rel_sequences(task["children"])



# IfcResources - IFC resources utility
# Copyright (C) 2021,2022 Dion Moult <dion@thinkmoult.com>, Yassine Oualid <yassine@sigmadimensions.com>
#
# This file is part of IfcResources.
#
# IfcResources is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcResources is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcResources.  If not, see <http://www.gnu.org/licenses/>.

import csv
import ifcopenshell
import ifcopenshell.api
import ifcopenshell.util.unit
import datetime


class Csv2Ifc:
    def __init__(self):
        self.csv = None
        self.file = None
        self.resources = []
        self.units = {}
        self.resource_map = {
            "CREW": "IfcCrewResource",
            "LABOR": "IfcLaborResource",
            "EQUIPMENT": "IfcConstructionEquipmentResource",
            "SUBCONTRACTOR": "IfcSubContractResource",
            "MATERIAL": "IfcConstructionMaterialResource",
            "PRODUCT": "IfcConstructionProductResource",
        }

    def execute(self):
        self.parse_csv()
        self.create_ifc()

    def parse_csv(self):
        self.parents = {}
        self.headers = {}
        with open(self.csv, "r") as csv_file:
            reader = csv.reader(csv_file)
            for row in reader:
                if not row[0]:
                    continue
                if row[0] == "HIERARCHY":
                    for i, col in enumerate(row):
                        if not col:
                            continue
                        self.headers[col] = i
                    continue
                resource_data = self.get_row_resource_data(row)
                hierarchy_key = int(row[0])
                if hierarchy_key == 1:
                    self.resources.append(resource_data)
                else:
                    self.parents[hierarchy_key - 1]["children"].append(resource_data)
                self.parents[hierarchy_key] = resource_data

    def get_row_resource_data(self, row):
        name = row[self.headers["ACTIVITY/RESOURCE NAME"]]
        resource_class = self.resource_map[row[self.headers["TYPE"]]]
        base_cost_value = row[self.headers["COST"]]
        productivity = {}

        if resource_class in ["IfcConstructionEquipmentResource", "IfcLaborResource"]:
            output_ratio = row[self.headers["LABOR OUTPUT"]]
            if not output_ratio:
                output_ratio = row[self.headers["EQUIPMENT OUTPUT"]]
            if output_ratio:
                time_consumed = datetime.timedelta(minutes=float(output_ratio) * 60)
                time_consumed = ifcopenshell.util.date.datetime2ifc(time_consumed, "IfcDuration")

                productivity = {
                    "BaseQuantityConsumed": time_consumed,
                    "BaseQuantityProducedName": row[self.headers["QUANTITY NAME"]],
                    "BaseQuantityProducedValue": 1,
                }

        return {
            "Name": str(name).strip() if name else None,
            "Description": row[self.headers["DESCRIPTION"]],
            "class": resource_class,
            "BaseCostValue": float(base_cost_value) if base_cost_value else None,
            "Productivity": productivity,
            "children": [],
        }

    def create_ifc(self):
        if not self.file:
            self.create_boilerplate_ifc()
        self.create_resources(self.resources)

    def create_resources(self, resources, parent=None):
        for resource in resources:
            self.create_resource(resource, parent)

    def create_resource(self, resource, parent):
        if parent is None:
            resource["ifc"] = ifcopenshell.api.run("resource.add_resource", self.file, ifc_class=resource["class"])
        else:
            resource["ifc"] = ifcopenshell.api.run(
                "resource.add_resource", self.file, parent_resource=parent, ifc_class=resource["class"]
            )
        resource["ifc"].Name = resource["Name"]
        if resource.get("Description", None):
            resource["ifc"].Description = resource.get("Description")
        if resource["Productivity"]:
            pset = ifcopenshell.api.run("pset.add_pset", self.file, product=resource["ifc"], name="EPset_Productivity")
            ifcopenshell.api.run(
                "pset.edit_pset",
                self.file,
                pset=pset,
                properties=resource["Productivity"],
            )
        if resource["BaseCostValue"]:
            cost_value = ifcopenshell.api.run("cost.add_cost_value", self.file, parent=resource["ifc"])
            cost_value.AppliedValue = self.file.createIfcMonetaryMeasure(resource["BaseCostValue"])
        self.create_resources(resource["children"], resource["ifc"])

    def create_unit(self, symbol, unit_type):
        unit = self.units.get(symbol, None)
        if unit:
            return unit
        else:
            unit = ifcopenshell.api.run("unit.add_si_unit", self.file, unit_type=unit_type)
        # unit = self.file.createIfcContextDependentUnit(
        #     self.file.createIfcDimensionalExponents(0, 0, 0, 0, 0, 0, 0), "USERDEFINED", symbol
        # )
        self.units[symbol] = unit
        return unit

    def create_boilerplate_ifc(self):
        self.file = ifcopenshell.file(schema="IFC4")
        ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcProject")
        ifcopenshell.api.run("unit.assign_unit", self.file)



print("In module products __package__, __name__ ==", __package__)
print(__name__)
import sys

print("In module products sys.path[0], __package__ ==", sys.path[0], __package__)
sys.path.append(sys.path[0])

import sqlite3
import math
import datetime
from datetime import timedelta
import ifcopenshell
import ifcopenshell.api
import ifcopenshell.util.date
from .wpattern import AstaCalendarWorkPattern
from .common import ScheduleIfcGenerator
import time


list_of_tables = [
    "BAR",
    "EXPANDED_TASK",
    "PERMANENT_RESOURCE",
    "CONSUMABLE_RESOURCE",
    "TASK",
    "PROJECT_SUMMARY",
    "WBS_ENTRY",
]


class PP2Ifc:
    def __init__(self):

        self.pp = None
        self.file = None
        self.work_plan = None
        self.project = {}
        self.calendars = {}
        self.wbs = {}
        self.root_activites = []
        self.activities = {}
        self.relationships = {}
        self.resources = {}
        self.output = None

        self.relationship_map = {0: "FINISH_START", 1: "FINISH_FINISH", 2: "START_START", 3: "START_FINISH"}

    def get_json(self, table_name):
        self.cur.execute("select * from " + table_name)
        r = [dict((self.cur.description[i][0], value) for i, value in enumerate(row)) for row in self.cur.fetchall()]
        return r

    def get_json_with_filter(self, table_name, attr_name, attr_value):
        self.cur.execute("select * from " + table_name + " where " + attr_name + " = " + str(attr_value))
        r = [dict((self.cur.description[i][0], value) for i, value in enumerate(row)) for row in self.cur.fetchall()]
        return r

    def execute(self):
        self.con = sqlite3.connect(self.pp)
        self.cur = self.con.cursor()
        self.parse_pp()
        settings = {
            "work_plan": self.work_plan,
            "project": self.project,
            "calendars": self.calendars,
            "wbs": self.wbs,
            "root_activities": self.root_activites,
            "activities": self.activities,
            "relationships": self.relationships,
            "resources": self.resources,
        }
        start = time.time()
        ifcCreator = ScheduleIfcGenerator(self.file, self.output, settings)
        end = time.time()

        ifcCreator.create_ifc()
        end2 = time.time()
        print("Parsing time is", end - start)
        print("IFC Creation took", end2 - end)
        print("Overall Time", end2 - start)
        # self.create_ifc()

    def parse_pp(self):

        project = self.get_json("PROJECT_SUMMARY")[0]
        self.project["Name"] = project["SHORT_NAME"]
        self.parse_calendar_pp()
        self.parse_bar()
        self.parse_relationship_pp(project)

    def parse_calendar_pp(self):
        calendars = self.get_json("CALENDAR")
        wp_data = self.get_json("WORK_PATTERN")
        work_types = self.get_json("EXCEPTIONN")
        work_type_ids = [wt["ID"] for wt in work_types if wt["EXCEPTION_TYPE"] == 0]

        for calendar in calendars:
            calendar_id = calendar["ID"]
            calendar_wp = calendar["DOMINANT_WORK_PATTERN"]
            wp_data = self.get_json_with_filter("WORK_PATTERN", "ID", calendar_wp)
            wp = AstaCalendarWorkPattern(wp_data[0]["SHIFTS"], work_type_ids)
            exceptions = {}
            timex = []
            for times in wp.dict_wp:
                work_times = timedelta(hours=0)
                for daily in times["WorkTimes"]:
                    fin = daily["Finish"]
                    strt = daily["Start"]
                    work_times += timedelta(hours=fin.hour, minutes=fin.minute) - timedelta(
                        hours=strt.hour, minutes=strt.minute
                    )
                timex.append(work_times.total_seconds() / (60 * 60))

            self.calendars[calendar_id] = {
                "Name": calendar["NAME"],
                "Type": "NOTDEFINED",
                "HoursPerDay": max(timex),
                "StandardWorkWeek": wp.dict_wp,
                "HolidayOrExceptions": exceptions,
            }
            # print(self.calendars[calendar_id])

    def parse_bar(self):
        bars = self.get_json("BAR")
        expanded_tasks = {t["BAR"]: t for t in self.get_json("EXPANDED_TASK")}
        tasks = {t["BAR"]: t for t in self.get_json("TASK")}
        milestones = {m["BAR"]: m for m in self.get_json("MILESTONE")}

        schedule_task = None

        bar_tasks = {}
        wbs_activities = {}

        for bar in bars:
            extra_type = None
            extra_data = None
            bar_id = None

            if bar["ID"] in expanded_tasks:
                extra_type = "EXPANDED_TASK"
                extra_data = expanded_tasks[bar["ID"]]
            elif bar["ID"] in tasks:
                extra_type = "TASK"
                extra_data = tasks[bar["ID"]]
            elif bar["ID"] in milestones:
                extra_type = "MILESTONE"
                extra_data = milestones[bar["ID"]]
            else:
                # Not sure where this might be the case
                continue

            name = bar["NAME"]
            if extra_data["NAME"]:
                name += f" - {extra_data['NAME']}"
            bar_tasks[extra_data["ID"]] = {"bar": bar, "extra_type": extra_type, "extra_data": extra_data}

            if extra_type == "EXPANDED_TASK":
                self.wbs[extra_data["ID"]] = {
                    "Name": name,
                    "Code": extra_data["ID"],
                    "ParentObjectId": bar["EXPANDED_TASK"] if bar["EXPANDED_TASK"] > 0 else None,
                    "ifc": None,
                    "rel": None,
                    "activities": [],
                }
            else:
                if "PLANNED_DURATION" in extra_data.keys():
                    activity_duration = float(
                        extra_data["PLANNED_DURATION"].split(",")[-2].replace("<", "").replace(">", "")
                    )
                elif "GIVEN_DURATION" in extra_data.keys():
                    activity_duration = float(
                        extra_data["GIVEN_DURATION"].split(",")[-2].replace("<", "").replace(">", "")
                    )
                else:
                    activity_duration = 0.0

                self.activities[extra_data["ID"]] = {
                    "Name": name,
                    "Identification": extra_data["ID"],
                    "StartDate": datetime.datetime.fromisoformat(extra_data["LINKABLE_START"]),
                    "FinishDate": datetime.datetime.fromisoformat(extra_data["LINKABLE_FINISH"]),
                    "PlannedDuration": activity_duration,
                    "Status": "PLANNED",
                    "CalendarObjectId": extra_data["CALENDAR"],
                    "ifc": None,
                }
                wbs_activities.setdefault(bar["EXPANDED_TASK"], []).append(extra_data["ID"])

        for wbs, activities in wbs_activities.items():
            self.wbs[wbs]["activities"] = activities

    def parse_relationship_pp(self, project):
        relations = self.get_json("LINK")
        # print(relations)
        for relationship in relations:
            predecessor = relationship["START_TASK"]
            successor = relationship["END_TASK"]
            if predecessor not in self.activities:
                print(f"Linked predecessor task {predecessor} cannot be found.")
                continue
            if successor not in self.activities:
                print(f"Linked successor task {successor} cannot be found.")
                continue
            self.relationships[relationship["ID"]] = {
                "PredecessorActivity": predecessor,
                "SuccessorActivity": successor,
                "Type": self.relationship_map[relationship["LINK_KIND"]],
                "Lag": float(relationship["END_LAG_TIME"].split(",")[-2].replace("<", "").replace(">", "")),
            }



from xerparser.reader import Reader
import ifcopenshell
import ifcopenshell.api
import ifcopenshell.util.date
from datetime import datetime, timedelta, date


class ScheduleIfcGenerator:
    def __init__(self, file, output, settings):
        self.file = file
        self.work_plan = settings["work_plan"]
        self.project = settings["project"]
        self.calendars = settings["calendars"]
        self.wbs = settings["wbs"]
        self.root_activites = settings["root_activities"]
        self.activities = settings["activities"]
        self.relationships = settings["relationships"]
        self.resources = settings["resources"]
        self.output = output
        self.day_map = {
            "Monday": 1,
            "Tuesday": 2,
            "Wednesday": 3,
            "Thursday": 4,
            "Friday": 5,
            "Saturday": 6,
            "Sunday": 7,
        }

    def create_ifc(self):
        if not self.file:
            self.file = self.create_boilerplate_ifc()
        if not self.work_plan:
            self.work_plan = ifcopenshell.api.run("sequence.add_work_plan", self.file)
        work_schedule = self.create_work_schedule()
        self.create_calendars()
        self.create_tasks(work_schedule)
        self.create_rel_sequences()
        self.create_resources()
        if self.output:
            self.file.write(self.output)

    def create_work_schedule(self):
        return ifcopenshell.api.run(
            "sequence.add_work_schedule", self.file, name=self.project["Name"], work_plan=self.work_plan
        )

    def create_calendars(self):
        for calendar in self.calendars.values():
            calendar["ifc"] = ifcopenshell.api.run("sequence.add_work_calendar", self.file, name=calendar["Name"])
            self.process_working_week(calendar["StandardWorkWeek"], calendar["ifc"])
            self.process_exceptions(calendar.get("HolidayOrExceptions"), calendar["ifc"])

    def process_working_week(self, week, calendar):
        for day in week:
            if day["ifc"] or not day.get("WorkTimes"):
                continue

            day["ifc"] = ifcopenshell.api.run(
                "sequence.add_work_time", self.file, work_calendar=calendar, time_type="WorkingTimes"
            )
            weekday_component = [self.day_map[day["DayOfWeek"]]]
            for day2 in week:
                if day["DayOfWeek"] == day2["DayOfWeek"]:
                    continue
                if day["WorkTimes"] == day2["WorkTimes"]:
                    weekday_component.append(self.day_map[day2["DayOfWeek"]])
                    # Don't process the next day, as we can group it
                    day2["ifc"] = day["ifc"]

            work_time_name = "Weekdays: {}".format(", ".join([str(c) for c in sorted(weekday_component)]))
            ifcopenshell.api.run(
                "sequence.edit_work_time",
                self.file,
                work_time=day["ifc"],
                attributes={"Name": work_time_name},
            )

            recurrence = ifcopenshell.api.run(
                "sequence.assign_recurrence_pattern", self.file, parent=day["ifc"], recurrence_type="WEEKLY"
            )
            ifcopenshell.api.run(
                "sequence.edit_recurrence_pattern",
                self.file,
                recurrence_pattern=recurrence,
                attributes={"WeekdayComponent": weekday_component},
            )
            for work_time in day["WorkTimes"]:
                ifcopenshell.api.run(
                    "sequence.add_time_period",
                    self.file,
                    recurrence_pattern=recurrence,
                    start_time=work_time["Start"],
                    end_time=work_time["Finish"],
                )

    def process_exceptions(self, exceptions, calendar):
        if exceptions:
            for year, year_data in exceptions.items():
                for month, month_data in year_data.items():
                    if month_data["FullDay"]:
                        self.process_full_day_exceptions(year, month, month_data, calendar)
                    if month_data["WorkTime"]:
                        self.process_work_time_exceptions(year, month, month_data, calendar)

    def process_full_day_exceptions(self, year, month, month_data, calendar):
        work_time = ifcopenshell.api.run(
            "sequence.add_work_time", self.file, work_calendar=calendar, time_type="ExceptionTimes"
        )
        ifcopenshell.api.run(
            "sequence.edit_work_time",
            self.file,
            work_time=work_time,
            attributes={
                "Name": f"{year}-{month}",
                "Start": date(year, 1, 1),
                "Finish": date(year, 12, 31),
            },
        )
        recurrence = ifcopenshell.api.run(
            "sequence.assign_recurrence_pattern",
            self.file,
            parent=work_time,
            recurrence_type="YEARLY_BY_DAY_OF_MONTH",
        )
        ifcopenshell.api.run(
            "sequence.edit_recurrence_pattern",
            self.file,
            recurrence_pattern=recurrence,
            attributes={"DayComponent": month_data["FullDay"], "MonthComponent": [month]},
        )

    def process_work_time_exceptions(self, year, month, month_data, calendar):
        for day in month_data["WorkTime"]:
            if day["ifc"]:
                continue

            day["ifc"] = ifcopenshell.api.run(
                "sequence.add_work_time", self.file, work_calendar=calendar, time_type="ExceptionTimes"
            )

            day_component = [day["Day"]]
            for day2 in month_data["WorkTime"]:
                if day["Day"] == day2["Day"]:
                    continue
                if day["WorkTimes"] == day2["WorkTimes"]:
                    day_component.append(day2["Day"])
                    # Don't process the next day, as we can group it
                    day2["ifc"] = day["ifc"]

            ifcopenshell.api.run(
                "sequence.edit_work_time",
                self.file,
                work_time=day["ifc"],
                attributes={
                    "Name": "{}-{}-{}".format(year, month, ", ".join([str(d) for d in day_component])),
                    "Start": date(year, 1, 1),
                    "Finish": date(year, 12, 31),
                },
            )
            recurrence = ifcopenshell.api.run(
                "sequence.assign_recurrence_pattern",
                self.file,
                parent=day["ifc"],
                recurrence_type="YEARLY_BY_DAY_OF_MONTH",
            )
            ifcopenshell.api.run(
                "sequence.edit_recurrence_pattern",
                self.file,
                recurrence_pattern=recurrence,
                attributes={"DayComponent": day_component, "MonthComponent": [month]},
            )
            for work_time in day["WorkTimes"]:
                ifcopenshell.api.run(
                    "sequence.add_time_period",
                    self.file,
                    recurrence_pattern=recurrence,
                    start_time=work_time["Start"],
                    end_time=work_time["Finish"],
                )

    def create_tasks(self, work_schedule):
        for wbs in self.wbs.values():
            self.create_task_from_wbs(wbs, work_schedule)
        for activity_id in self.root_activites:
            self.create_task_from_activity(self.activities[activity_id], None, work_schedule)

    def create_task_from_wbs(self, wbs, work_schedule):
        if not self.wbs.get(wbs["ParentObjectId"]):
            wbs["ParentObjectId"] = None
        wbs["ifc"] = ifcopenshell.api.run(
            "sequence.add_task",
            self.file,
            work_schedule=None if wbs["ParentObjectId"] else work_schedule,
            parent_task=self.wbs[wbs["ParentObjectId"]]["ifc"] if wbs["ParentObjectId"] else None,
        )
        identification = wbs["Code"]
        if wbs["ParentObjectId"]:
            if self.wbs[wbs["ParentObjectId"]]["ifc"]:
                identification = str(self.wbs[wbs["ParentObjectId"]]["ifc"].Identification) + "." + str(wbs["Code"])
        ifcopenshell.api.run(
            "sequence.edit_task",
            self.file,
            task=wbs["ifc"],
            attributes={"Name": wbs["Name"], "Identification": str(identification)},
        )
        for activity_id in wbs["activities"]:
            self.create_task_from_activity(self.activities[activity_id], wbs, None)

    def create_task_from_activity(self, activity, wbs, work_schedule):
        activity["ifc"] = ifcopenshell.api.run(
            "sequence.add_task",
            self.file,
            work_schedule=None if wbs else work_schedule,
            parent_task=wbs["ifc"] if wbs else None,
        )
        ifcopenshell.api.run(
            "sequence.edit_task",
            self.file,
            task=activity["ifc"],
            attributes={
                "Name": activity["Name"],
                "Identification": str(activity["Identification"]),
                "Status": activity["Status"],
                "IsMilestone": activity["StartDate"] == activity["FinishDate"],
                "PredefinedType": "CONSTRUCTION",
            },
        )
        task_time = ifcopenshell.api.run("sequence.add_task_time", self.file, task=activity["ifc"])
        calendar = self.calendars[activity["CalendarObjectId"]]
        # print(calendar, self.calendars)
        # Seems intermittently crashy - can we investigate for larger files?
        ifcopenshell.api.run(
            "control.assign_control",
            self.file,
            **{
                "relating_control": calendar["ifc"],
                "related_object": activity["ifc"],
            },
        )
        ifcopenshell.api.run(
            "sequence.edit_task_time",
            self.file,
            task_time=task_time,
            attributes={
                "ScheduleStart": activity["StartDate"],
                "ScheduleFinish": activity["FinishDate"],
                "DurationType": "WORKTIME" if activity["PlannedDuration"] else None,
                "ScheduleDuration": timedelta(
                    days=float(activity["PlannedDuration"]) / float(calendar["HoursPerDay"] or 8)
                )
                or None
                if activity["PlannedDuration"]
                else None,
            },
        )

    def create_rel_sequences(self):
        self.sequence_type_map = {
            "Start to Start": "START_START",
            "Start to Finish": "START_FINISH",
            "Finish to Start": "FINISH_START",
            "Finish to Finish": "FINISH_FINISH",
        }
        for relationship in self.relationships.values():
            rel_sequence = ifcopenshell.api.run(
                "sequence.assign_sequence",
                self.file,
                relating_process=self.activities[relationship["PredecessorActivity"]]["ifc"],
                related_process=self.activities[relationship["SuccessorActivity"]]["ifc"],
            )
            ifcopenshell.api.run(
                "sequence.edit_sequence",
                self.file,
                rel_sequence=rel_sequence,
                attributes={"SequenceType": relationship["Type"]},
            )
            lag = float(relationship["Lag"])
            if lag:
                calendar = self.calendars[self.activities[relationship["PredecessorActivity"]]["CalendarObjectId"]]
                ifcopenshell.api.run(
                    "sequence.assign_lag_time",
                    self.file,
                    rel_sequence=rel_sequence,
                    lag_value=timedelta(days=lag / float(calendar["HoursPerDay"] or 8)),
                    duration_type="WORKTIME",
                )

    def create_resources(self):
        # print("Resources", self.resources)
        if self.resources:
            for id, resource in self.resources.items():

                parent = self.resources.get(resource.get("ParentObjectId"))
                if parent:
                    if not parent.get("ifc"):
                        parent["ifc"] = ifcopenshell.api.run(
                            "resource.add_resource",
                            self.file,
                            **{"ifc_class": "IfcCrewResource", "name": parent["Name"]},
                        )
                if parent:
                    resource["ifc"] = ifcopenshell.api.run(
                        "resource.add_resource",
                        self.file,
                        **{
                            "parent_resource": parent["ifc"] if parent else None,
                            "ifc_class": "IfcCrewResource",
                            "name": resource["Name"],
                        },
                    )
                else:
                    resource["ifc"] = ifcopenshell.api.run(
                        "resource.add_resource", self.file, **{"ifc_class": "IfcCrewResource", "name": resource["Name"]}
                    )
            print(self.resources)

    def create_boilerplate_ifc(self):
        self.file = ifcopenshell.file(schema="IFC4")
        self.work_plan = self.file.create_entity("IfcWorkPlan")



# Ifc4D - IFC scheduling utility
# Copyright (C) 2020-2024 Dion Moult <dion@thinkmoult.com>
#
# This file is part of Ifc4D.
#
# Ifc4D is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ifc4D is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Ifc4D.  If not, see <http://www.gnu.org/licenses/>.

__version__ = version = "0.7.9"



# Ifc4D - IFC scheduling utility
# Copyright (C) 2022 Dion Moult <dion@thinkmoult.com>
#
# This file is part of Ifc4D.
#
# Ifc4D is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ifc4D is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Ifc4D.  If not, see <http://www.gnu.org/licenses/>.

import uuid
import datetime
import ifcopenshell
import ifcopenshell.util.date
import xml.etree.ElementTree as ET


class Ifc2Msp:
    def __init__(self):
        self.xml = None
        self.file = None
        self.id = 0
        self.id_map = {}
        self.element_map = {}
        self.hours_per_day = 8
        self.project_tasks = []
        self.start_dates = []
        self.base_calendar = None
        # P6XML does not understand patterns in holiday dates. Instead, it
        # expects a list of every day that is a holiday between a date range.
        # This allows you to specify the date range where the export should
        # calculate holidays for.
        self.holiday_start_date = None
        self.holiday_finish_date = None
        self.work_schedule = None

    def execute(self):
        self.root = ET.Element("Project")
        self.root.attrib["xmlns"] = "http://schemas.microsoft.com/project"

        self.schedule = self.file.by_type("IfcWorkSchedule")[0]

        self.create_project()
        self.create_calendars()
        self.create_tasks()

        self.set_start_date()
        self.set_base_calendar()

        tree = ET.ElementTree(self.root)
        tree.write(self.xml, encoding="utf-8", xml_declaration=True)

    def create_project(self):
        # Which version should we pick? I don't know. Hardcode as 14 until it breaks :)
        # Values are: 12=Project 2007, 24=Project 2010
        ET.SubElement(self.root, "SaveVersion").text = "14"
        element = self.file.by_type("IfcProject")[0]
        self.link_element(element, self.root)
        ET.SubElement(self.root, "GUID").text = str(uuid.UUID(ifcopenshell.guid.expand(element.GlobalId))).upper()
        ET.SubElement(self.root, "Name").text = element.Name or "X"
        ET.SubElement(self.root, "Title").text = element.LongName or "Unnamed"
        # This is mandatory, but we don't use it
        ET.SubElement(self.root, "CurrencyCode").text = "USD"
        ET.SubElement(self.root, "DurationFormat").text = "7"
        ET.SubElement(self.root, "MinutesPerDay").text = "480"

    def set_start_date(self):
        ET.SubElement(self.root, "ScheduleFromStart").text = "1"
        if self.start_dates:
            # Set to midnight as a workaround. If I don't, the first duration seems to stick to zero.
            ET.SubElement(self.root, "StartDate").text = sorted(self.start_dates)[0].split("T")[0] + "T00:00:00"

    def set_base_calendar(self):
        if self.base_calendar:
            ET.SubElement(self.root, "CalendarUID").text = self.id_map[self.base_calendar]

    def create_calendars(self):
        el = ET.SubElement(self.root, "Calendars")
        for calendar in self.file.by_type("IfcWorkCalendar"):
            self.create_calendar(calendar, el)

    def create_calendar(self, calendar, parent):
        el = ET.SubElement(parent, "Calendar")
        self.link_element(calendar, el)
        ET.SubElement(el, "Name").text = calendar.Name or "Unnamed"
        ET.SubElement(el, "IsBaseCalendar").text = "1"

        working_week = self.auto_detect_working_week(calendar)

        if not working_week:
            working_week = self.get_default_working_week()

        week_days = ET.SubElement(el, "WeekDays")

        day_map = {
            "Sunday": "1",
            "Monday": "2",
            "Tuesday": "3",
            "Wednesday": "4",
            "Thursday": "5",
            "Friday": "6",
            "Saturday": "7",
        }

        for day in ["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]:
            time_periods = working_week.get(day, None)
            if time_periods:
                week_day = ET.SubElement(week_days, "WeekDay")
                ET.SubElement(week_day, "DayType").text = day_map[day]
                ET.SubElement(week_day, "DayWorking").text = "1"
                working_times = ET.SubElement(week_day, "WorkingTimes")
                for time_period in time_periods:
                    working_time = ET.SubElement(working_times, "WorkingTime")
                    ET.SubElement(working_time, "FromTime").text = time_period["Start"]
                    ET.SubElement(working_time, "ToTime").text = time_period["Finish"]
            else:
                week_day = ET.SubElement(week_days, "WeekDay")
                ET.SubElement(week_day, "DayType").text = day_map[day]
                ET.SubElement(week_day, "DayWorking").text = "0"

        # TODO Exceptions not yet implemented
        return

        holidays = ET.SubElement(el, "Exceptions")
        for holiday in self.auto_detect_holidays(calendar):
            el = ET.SubElement(holidays, "Exception")
            # Recurrence is supported but I haven't implemented it yet
            holiday = ifcopenshell.util.date.datetime2ifc(holiday, "IfcDateTime")
            time_period = ET.SubElement(el, "TimePeriod")
            ET.SubElement(time_period, "FromDate").text = holiday
            ET.SubElement(time_period, "ToDate").text = holiday

    def auto_detect_working_week(self, calendar):
        # Microsoft Project XML only understands a working week. In other words,
        # it understands the equivalent of an IFC weekly recurring time period.
        # If you do not have a weekly recurring time period, we just give a
        # default working week.
        results = {}
        weekday_component_map = {
            1: "Monday",
            2: "Tuesday",
            3: "Wednesday",
            4: "Thursday",
            5: "Friday",
            6: "Saturday",
            7: "Sunday",
        }
        for working_time in calendar.WorkingTimes:
            if not working_time.RecurrencePattern or working_time.RecurrencePattern.RecurrenceType != "WEEKLY":
                continue

            if not working_time.RecurrencePattern.TimePeriods:
                time_periods = [{"Start": "09:00:00", "Finish": "17:00:00"}]
            else:
                time_periods = [
                    {"Start": p.StartTime, "Finish": p.EndTime} for p in working_time.RecurrencePattern.TimePeriods
                ]

            for component in working_time.RecurrencePattern.WeekdayComponent:
                results.setdefault(weekday_component_map[component], []).extend(time_periods)

        return results

    def get_default_working_week(self):
        results = {}
        # As a fallback, we work 5 days a week, 8 hours per day.
        for name in ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday"]:
            results[name] = [{"Start": "09:00:00", "Finish": "17:00:00"}]
        return results

    def auto_detect_holidays(self, calendar):
        results = []
        start = self.holiday_start_date
        finish = self.holiday_finish_date
        if finish < start:
            return results
        while start < finish:
            start += datetime.timedelta(days=1)
            if not ifcopenshell.util.sequence.is_working_day(start, calendar):
                results.append(start)
        return results

    def create_tasks(self):
        self.tasks = ET.SubElement(self.root, "Tasks")
        for task in self.get_subtasks(self.work_schedule):
            self.create_task(task)

    def create_task(self, task, level=0):
        el = ET.SubElement(self.tasks, "Task")
        self.link_element(task, el)
        ET.SubElement(el, "Name").text = task.Name or "Unnamed"
        ET.SubElement(el, "OutlineNumber").text = task.Identification or "X"
        ET.SubElement(el, "OutlineLevel").text = str(level)
        ET.SubElement(el, "DurationFormat").text = "53"
        if task.TaskTime:
            if task.TaskTime.ScheduleDuration:
                duration = ifcopenshell.util.date.ifc2datetime(task.TaskTime.ScheduleDuration)
                ET.SubElement(el, "Duration").text = f"PT{duration.days * 8}H0M0S"
            if task.TaskTime.ScheduleStart:
                ET.SubElement(el, "Start").text = task.TaskTime.ScheduleStart
                self.start_dates.append(task.TaskTime.ScheduleStart)
            if task.TaskTime.ScheduleFinish:
                ET.SubElement(el, "Finish").text = task.TaskTime.ScheduleFinish
            ET.SubElement(el, "Critical").text = "1" if task.TaskTime.IsCritical else "0"

            data_map = {
                "EarlyStart": task.TaskTime.EarlyStart,
                "EarlyFinish": task.TaskTime.EarlyFinish,
                "LateStart": task.TaskTime.LateStart,
                "LateFinish": task.TaskTime.LateFinish,
            }
            for key, value in data_map.items():
                if value:
                    ET.SubElement(el, key).text = value

        calendar = ifcopenshell.util.sequence.derive_calendar(task)
        if calendar:
            ET.SubElement(el, "CalendarUID").text = self.id_map[calendar]
            if level == 0:
                # This must be the base calendar
                self.base_calendar = calendar

        for rel in task.IsSuccessorFrom or []:
            predecessor = rel.RelatingProcess
            predecessor_link = ET.SubElement(el, "PredecessorLink")
            self.link_element(rel, predecessor_link)
            if rel.TimeLag and rel.TimeLag.LagValue:
                duration = ifcopenshell.util.date.ifc2datetime(rel.TimeLag.LagValue.wrappedValue)
                # Lag time is expressed in tenths of a minute. Seriously, Microsoft?
                ET.SubElement(predecessor_link, "LinkLag").text = str(duration.days * self.hours_per_day * 60 * 10)
            else:
                ET.SubElement(predecessor_link, "LinkLag").text = "0"
            ET.SubElement(predecessor_link, "PredecessorUID").text = str(predecessor.id())
            ET.SubElement(predecessor_link, "Type").text = {
                "START_START": "3",
                "START_FINISH": "2",
                "FINISH_START": "1",
                "FINISH_FINISH": "0",
                "NOTDEFINED": "1",
                "USERDEFINED": "1",
                None: "1",
            }[rel.SequenceType]

        for subtask in self.get_subtasks(task):
            self.create_task(subtask, level=level + 1)

    def get_subtasks(self, element):
        tasks = []
        if element.is_a("IfcWorkSchedule"):
            for rel in element.Controls or []:
                tasks.extend([e for e in rel.RelatedObjects if e.is_a("IfcTask")])
        elif element.is_a("IfcTask"):
            for rel in element.IsNestedBy or []:
                tasks.extend([e for e in rel.RelatedObjects if e.is_a("IfcTask")])
        return tasks

    def link_element(self, element, el):
        ET.SubElement(el, "UID").text = str(element.id())
        ET.SubElement(el, "GUID").text = str(uuid.UUID(ifcopenshell.guid.expand(element.GlobalId))).upper()
        self.id_map[element] = str(element.id())
        self.element_map[element] = el



# Ifc4D - IFC scheduling utility
# Copyright (C) 2021, 2022 Dion Moult <dion@thinkmoult.com>
#
# This file is part of Ifc4D.
#
# Ifc4D is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ifc4D is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Ifc4D.  If not, see <http://www.gnu.org/licenses/>.

import math
import datetime
import ifcopenshell
import ifcopenshell.api
import ifcopenshell.util.date
import xml.etree.ElementTree as ET
from .common import ScheduleIfcGenerator


class P62Ifc:
    def __init__(self):
        self.xml = None
        self.file = None
        self.work_plan = None
        self.project = {}
        self.calendars = {}
        self.wbs = {}
        self.root_activites = []
        self.activities = {}
        self.relationships = {}
        self.resources = {}
        self.output = None
        self.day_map = {
            "Monday": 1,
            "Tuesday": 2,
            "Wednesday": 3,
            "Thursday": 4,
            "Friday": 5,
            "Saturday": 6,
            "Sunday": 7,
        }
        self.sequence_type_map = {
            "Start to Start": "START_START",
            "Start to Finish": "START_FINISH",
            "Finish to Start": "FINISH_START",
            "Finish to Finish": "FINISH_FINISH",
        }
        self.resource_type_map = {
            "Labor": "LABOR",
            "Mat": "MATERIAL",
            "Equip": "EQUIPMENT",
            # https://www.primaverascheduling.com/category/primavera-resources/
            "Nonlabor": "EQUIPMENT",
        }

    def execute(self):
        import time

        start = time.time()
        print("Started")
        self.parse_xml()
        settings = {
            "work_plan": self.work_plan,
            "project": self.project,
            "calendars": self.calendars,
            "wbs": self.wbs,
            "root_activities": self.root_activites,
            "activities": self.activities,
            "relationships": self.relationships,
            "resources": self.resources,
        }
        ifcCreator = ScheduleIfcGenerator(self.file, self.output, settings)
        end = time.time()

        ifcCreator.create_ifc()
        end2 = time.time()
        print("Parsing time is", end - start)
        print("IFC Creation took", end2 - end)
        print("Overall Time", end2 - start)

    def parse_xml(self):
        tree = ET.parse(self.xml)
        root = tree.getroot()
        self.ns = {"pr": root.tag[1:].partition("}")[0]}
        project = root.find("pr:Project", self.ns)
        self.project["Name"] = project.findtext("pr:Name") or "Unnamed"
        self.parse_calendar_xml(root)
        self.parse_calendar_xml(project)
        self.parse_wbs_xml(project)
        self.parse_activity_xml(project)
        self.parse_relationship_xml(project)
        self.parse_resources_xml(root)

    def parse_calendar_xml(self, project):
        for calendar in project.findall("pr:Calendar", self.ns):
            calendar_id = calendar.find("pr:ObjectId", self.ns).text
            standard_work_week = []
            for standard_work_hour in calendar.find("pr:StandardWorkWeek", self.ns).findall(
                "pr:StandardWorkHours", self.ns
            ):
                work_times = []
                for work_time in standard_work_hour.findall("pr:WorkTime", self.ns):
                    if work_time.find("pr:Start", self.ns) is None:
                        continue
                    work_times.append(
                        {
                            "Start": datetime.time.fromisoformat(work_time.find("pr:Start", self.ns).text),
                            "Finish": datetime.time.fromisoformat(work_time.find("pr:Finish", self.ns).text),
                        }
                    )
                standard_work_week.append(
                    {
                        "DayOfWeek": standard_work_hour.find("pr:DayOfWeek", self.ns).text,
                        "WorkTimes": work_times,
                        "ifc": None,
                    }
                )
            exceptions = {}
            holiday_or_exceptions = calendar.find("pr:HolidayOrExceptions", self.ns)
            holiday_or_exception = []
            if holiday_or_exceptions is not None:
                holiday_or_exception = holiday_or_exceptions.findall("pr:HolidayOrException", self.ns)
            for exception in holiday_or_exception:
                d = datetime.datetime.fromisoformat(exception.find("pr:Date", self.ns).text).date()
                month = exceptions.setdefault(d.year, {}).setdefault(d.month, {})
                month.setdefault("FullDay", [])
                month.setdefault("WorkTime", [])
                work_times = []
                for work_time in exception.findall("pr:WorkTime", self.ns):
                    if work_time.find("pr:Start", self.ns) is None:
                        continue
                    work_times.append(
                        {
                            "Start": datetime.time.fromisoformat(work_time.find("pr:Start", self.ns).text),
                            "Finish": datetime.time.fromisoformat(work_time.find("pr:Finish", self.ns).text),
                        }
                    )
                if work_times:
                    exceptions[d.year][d.month]["WorkTime"].append({"Day": d.day, "WorkTimes": work_times, "ifc": None})
                else:
                    exceptions[d.year][d.month]["FullDay"].append(d.day)
            self.calendars[calendar_id] = {
                "Name": calendar.find("pr:Name", self.ns).text,
                "Type": calendar.find("pr:Type", self.ns).text,
                "HoursPerDay": calendar.find("pr:HoursPerDay", self.ns).text,
                "StandardWorkWeek": standard_work_week,
                "HolidayOrExceptions": exceptions,
            }

    def parse_wbs_xml(self, project):
        for wbs in project.findall("pr:WBS", self.ns):
            self.wbs[wbs.find("pr:ObjectId", self.ns).text] = {
                "Name": wbs.find("pr:Name", self.ns).text,
                "Code": wbs.find("pr:Code", self.ns).text,
                "ParentObjectId": wbs.find("pr:ParentObjectId", self.ns).text,
                "ifc": None,
                "rel": None,
                "activities": [],
            }

    def parse_activity_xml(self, project):
        for activity in project.findall("pr:Activity", self.ns):
            activity_type = activity.find("pr:Type", self.ns).text
            if activity_type == "Level of Effort":
                continue
            activity_id = activity.find("pr:ObjectId", self.ns).text
            wbs_id = activity.find("pr:WBSObjectId", self.ns).text
            if wbs_id:
                self.wbs[wbs_id]["activities"].append(activity_id)
            else:
                self.root_activites.append(activity_id)
            self.activities[activity_id] = {
                "Name": activity.find("pr:Name", self.ns).text,
                "Identification": activity.find("pr:Id", self.ns).text,
                "StartDate": datetime.datetime.fromisoformat(activity.find("pr:StartDate", self.ns).text),
                "FinishDate": datetime.datetime.fromisoformat(activity.find("pr:FinishDate", self.ns).text),
                "PlannedDuration": activity.find("pr:PlannedDuration", self.ns).text,
                "Status": activity.find("pr:Status", self.ns).text,
                "CalendarObjectId": activity.find("pr:CalendarObjectId", self.ns).text,
                "ifc": None,
            }

    def parse_relationship_xml(self, project):
        for relationship in project.findall("pr:Relationship", self.ns):
            predecessor = relationship.find("pr:PredecessorActivityObjectId", self.ns).text
            successor = relationship.find("pr:SuccessorActivityObjectId", self.ns).text
            if predecessor not in self.activities or successor not in self.activities:
                continue
            self.relationships[relationship.find("pr:ObjectId", self.ns).text] = {
                "PredecessorActivity": predecessor,
                "SuccessorActivity": successor,
                "Type": self.sequence_type_map[relationship.find("pr:Type", self.ns).text],
                "Lag": relationship.find("pr:Lag", self.ns).text,
            }

    def get_wbs(self, wbs):
        return {"Name": wbs.find("pr:Name", self.ns).text, "subtasks": []}

    def parse_resources_xml(self, project):
        resources = project.findall("pr:Resource", self.ns)
        for resource in resources:
            id = resource.find("pr:ObjectId", self.ns).text
            self.resources[id] = {
                "Name": resource.find("pr:Name", self.ns).text,
                "Code": resource.find("pr:Id", self.ns).text,
                "ParentObjectId": resource.find("pr:ParentObjectId", self.ns).text,
                "Type": self.resource_type_map[resource.find("pr:ResourceType", self.ns).text],
                "ifc": None,
                "rel": None,
            }
        print("Resource found", self.resources)



"""
This class parses the calendar work pattern and retruns a list
The list returned has a key DayOfWeek which takes a value Sunday to Saturday
for each day as a key there is a list of working times with the format 
{"Start": datetime.time, "Finish": datetime.time}

"""

import re
from datetime import time, datetime


class AstaCalendarWorkPattern:
    def get_keys(self, s):
        regex = r"\<(\".+?\")\>(\w|\d|)+?"

        matches = re.finditer(regex, s)
        matcs = []
        for matchNum, match in enumerate(matches, start=1):
            matcs.append(match.group(1))
        return matcs

    def get_values(self, s):
        rx2 = r"<\"[^<>]+\">"
        data = re.split("<[^<>]+>", s)
        return data

    def __init__(self, string, work_type_ids):
        self.string = string
        self.Days = {
            "en": ["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"],
            "sv": ["Söndag", "Måndag", "Tisdag", "Onsdag", "Torsdag", "Fredag", "Lördag"],
            "de": ["Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"],
        }
        self.keys = self.get_keys(string)
        self.values = self.get_values(string)
        self.dict_wp = []
        for d, m in zip(self.values[1:], self.keys):
            splt_data = d.strip().split(",")
            workhours = []
            if len(splt_data) >= 7:
                number_of_work_slots = splt_data[1]
                for i in range(int(number_of_work_slots)):
                    if int(splt_data[2 + i * 3]) in work_type_ids:
                        st1_1 = datetime.strptime(splt_data[2 + i * 3 + 1].ljust(5, "0"), "%H%M%S")
                        st1_2 = datetime.strptime(splt_data[2 + i * 3 + 2].ljust(5, "0"), "%H%M%S")
                        st = {
                            "Start": time(st1_1.hour, st1_1.minute),
                            "Finish": time(st1_2.hour, st1_2.minute),
                            "ifc": None,
                        }
                        workhours.append(st)

            self.dict_wp.append({"DayOfWeek": m.replace('"', ""), "WorkTimes": workhours, "ifc": None})

            # Translate day names to english
            def translate_days(days, wp):
                for index, day in enumerate(days["en"]):
                    for lang in days.keys():
                        if lang == "en":
                            continue
                        if wp["DayOfWeek"] == days[lang][index]:
                            wp["DayOfWeek"] = days["en"][index]
                            return
            translate_days(self.Days, self.dict_wp[-1])

        for day in self.Days["en"]:
            if not len(list(filter(lambda d: d["DayOfWeek"] == day, self.dict_wp))) > 0:
                print("MISSING", day)
                self.dict_wp.append({"DayOfWeek": day, "WorkTimes": [], "ifc": None})



# Ifc4D - IFC scheduling utility
# Copyright (C) 2022 Dion Moult <dion@thinkmoult.com>
#
# This file is part of Ifc4D.
#
# Ifc4D is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ifc4D is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Ifc4D.  If not, see <http://www.gnu.org/licenses/>.

import uuid
import datetime
import ifcopenshell
import ifcopenshell.util.date
import xml.etree.ElementTree as ET
from .common import ScheduleIfcGenerator


class Ifc2P6:
    def __init__(self):
        self.xml = None
        self.file = None
        self.id = 0
        self.id_map = {}
        self.element_map = {}
        self.hours_per_day = 8
        self.project_tasks = []
        self.starts = []
        self.finishes = []
        # P6XML does not understand patterns in holiday dates. Instead, it
        # expects a list of every day that is a holiday between a date range.
        # This allows you to specify the date range where the export should
        # calculate holidays for.
        self.holiday_start_date = None
        self.holiday_finish_date = None

    def execute(self):
        self.root = ET.Element("APIBusinessObjects")
        self.root.attrib["xmlns"] = "http://xmlns.oracle.com/Primavera/P6Professional/V18.8/API/BusinessObjects"
        self.root.attrib["xmlns:xsi"] = "http://www.w3.org/2001/XMLSchema-instance"
        self.root.attrib[
            "xsi:schemaLocation"
        ] = "http://xmlns.oracle.com/Primavera/P6Professional/V18.8/API/BusinessObjects http://xmlns.oracle.com/Primavera/P6Professional/V18.8/API/p6apibo.xsd"

        self.schedule = self.file.by_type("IfcWorkSchedule")[0]

        # It seems as though OBSes aren't necessary and cause permission issues on P6 enterprise setups
        # self.create_obs()
        self.create_projects()

        tree = ET.ElementTree(self.root)
        tree.write(self.xml, encoding="utf-8", xml_declaration=True)

    def create_obs(self):
        element = self.file.by_type("IfcProject")[0]
        obs = ET.SubElement(self.root, "OBS")
        self.link_element(element, obs)
        ET.SubElement(obs, "Id").text = element.Name or "X"
        ET.SubElement(obs, "Name").text = element.LongName or "Unnamed"

    def create_calendars(self, work_schedule):
        for calendar in self.file.by_type("IfcWorkCalendar"):
            self.create_calendar(calendar, work_schedule)

    def create_calendar(self, calendar, work_schedule):
        el = ET.SubElement(self.element_map[work_schedule], "Calendar")
        ET.SubElement(el, "Name").text = calendar.Name or "Unnamed"
        self.link_element(calendar, el)
        ET.SubElement(el, "BaseCalendarObjectId")
        ET.SubElement(el, "HoursPerDay").text = "8"
        ET.SubElement(el, "HoursPerMonth").text = "172"
        ET.SubElement(el, "HoursPerWeek").text = "40"
        ET.SubElement(el, "HoursPerYear").text = "2000"
        ET.SubElement(el, "IsDefault").text = "0"
        ET.SubElement(el, "IsPersonal").text = "0"
        ET.SubElement(el, "ProjectObjectId").text = self.id_map[work_schedule]
        ET.SubElement(el, "Type").text = "Project"

        working_week = self.auto_detect_working_week(calendar)

        if not working_week:
            working_week = self.get_default_working_week()

        work_week = ET.SubElement(el, "StandardWorkWeek")

        for day in ["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]:
            work_hours = ET.SubElement(work_week, "StandardWorkHours")
            ET.SubElement(work_hours, "DayOfWeek").text = day
            if day in working_week:
                print(working_week)
                for time_period in working_week[day]:
                    work_time = ET.SubElement(work_hours, "WorkTime")
                    ET.SubElement(work_time, "Start").text = time_period["Start"]
                    ET.SubElement(work_time, "Finish").text = time_period["Finish"]
            else:
                ET.SubElement(work_hours, "WorkTime")

        holidays = ET.SubElement(el, "HolidayOrExceptions")
        for holiday in self.auto_detect_holidays(calendar):
            el = ET.SubElement(holidays, "HolidayOrException")
            ET.SubElement(el, "Date").text = ifcopenshell.util.date.datetime2ifc(holiday, "IfcDateTime")
            ET.SubElement(el, "WorkTime")

    def auto_detect_working_week(self, calendar):
        # P6 XML only understands a working week. In other words, it understands
        # the equivalent of an IFC weekly recurring time period. If you do not
        # have a weekly recurring time period, we just give a default working
        # week.
        results = {}
        weekday_component_map = {
            1: "Monday",
            2: "Tuesday",
            3: "Wednesday",
            4: "Thursday",
            5: "Friday",
            6: "Saturday",
            7: "Sunday",
        }
        for working_time in calendar.WorkingTimes or []:
            if not working_time.RecurrencePattern or working_time.RecurrencePattern.RecurrenceType != "WEEKLY":
                continue

            if not working_time.RecurrencePattern.TimePeriods:
                time_periods = [{"Start": "09:00:00", "Finish": "17:00:00"}]
            else:
                time_periods = [
                    {"Start": p.StartTime, "Finish": p.EndTime} for p in working_time.RecurrencePattern.TimePeriods
                ]

            for component in working_time.RecurrencePattern.WeekdayComponent:
                results.setdefault(weekday_component_map[component], []).extend(time_periods)

        return results

    def get_default_working_week(self):
        results = {}
        # As a fallback, we work 5 days a week, 8 hours per day.
        for name in ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday"]:
            results[name] = [{"Start": "09:00:00", "Finish": "17:00:00"}]
        return results

    def auto_detect_holidays(self, calendar):
        results = []
        start = self.holiday_start_date
        finish = self.holiday_finish_date
        if finish < start:
            return results
        while start < finish:
            start += datetime.timedelta(days=1)
            if not ifcopenshell.util.sequence.is_working_day(start, calendar):
                results.append(start)
        return results

    def create_projects(self):
        for work_schedule in self.file.by_type("IfcWorkSchedule"):
            self.create_project(work_schedule)

    def create_project(self, work_schedule):
        project = ET.SubElement(self.root, "Project")
        self.link_element(work_schedule, project)
        activity_calendar = ET.SubElement(project, "ActivityDefaultCalendarObjectId")
        ET.SubElement(project, "Name").text = work_schedule.Name or "Unnamed"
        ET.SubElement(project, "Id").text = work_schedule.Identification or "X"
        data_date = ET.SubElement(project, "DataDate")
        planned_start_date = ET.SubElement(project, "PlannedStartDate")
        # ET.SubElement(project, "OBSObjectId").text = self.id_map[self.file.by_type("IfcProject")[0]]

        # Note: order matters for P6 enterprise imports.
        # Root elements like Name must come first. Then calendar, WBS, activities, and relationships.
        # Changing the order will result in import failures (but not for all P6 versions, it seems?)
        self.create_calendars(work_schedule)
        self.project_tasks = []
        self.activities = []
        self.create_wbs(self.get_subtasks(work_schedule), work_schedule)
        self.create_activities()
        self.create_relationships(self.project_tasks, work_schedule)

        activity_calendar.text = self.id_map[self.file.by_type("IfcWorkCalendar")[0]]
        start_date = sorted(self.starts)[0]
        data_date.text = start_date
        planned_start_date.text = start_date

    def create_wbs(self, tasks, work_schedule, parent=None):
        self.project_tasks.extend(tasks)
        for task in tasks:
            subtasks = self.get_subtasks(task)
            if not subtasks and task.TaskTime:
                self.activities.append({"task": task, "work_schedule": work_schedule, "parent": parent})
                continue
            wbs = ET.SubElement(self.element_map[work_schedule], "WBS")
            self.link_element(task, wbs)
            ET.SubElement(wbs, "Code").text = task.Identification or "X"
            ET.SubElement(wbs, "Name").text = task.Name or "Unnamed"
            parent_object_id = ET.SubElement(wbs, "ParentObjectId")
            if parent:
                parent_object_id.text = self.id_map[parent]
            ET.SubElement(wbs, "ProjectObjectId").text = self.id_map[work_schedule]
            self.create_wbs(subtasks, work_schedule, parent=task)

    def create_activities(self):
        for activity in self.activities:
            self.create_activity(activity["task"], activity["work_schedule"], parent=activity["parent"])

    def create_activity(self, task, work_schedule, parent=None):
        activity = ET.SubElement(self.element_map[work_schedule], "Activity")
        self.link_element(task, activity)
        ET.SubElement(activity, "Id").text = task.Identification or "X"
        ET.SubElement(activity, "Name").text = task.Name or "Unnamed"
        if parent:
            ET.SubElement(activity, "WBSObjectId").text = self.id_map[parent]
        if task.TaskTime.ScheduleDuration:
            duration = ifcopenshell.util.date.ifc2datetime(task.TaskTime.ScheduleDuration)
            ET.SubElement(activity, "PlannedDuration").text = str(duration.days * self.hours_per_day)
            ET.SubElement(activity, "RemainingDuration").text = str(duration.days * self.hours_per_day)
        if task.TaskTime.ScheduleStart:
            ET.SubElement(activity, "PlannedStartDate").text = task.TaskTime.ScheduleStart
            ET.SubElement(activity, "StartDate").text = task.TaskTime.ScheduleStart
            self.starts.append(task.TaskTime.ScheduleStart)
        if task.TaskTime.ScheduleFinish:
            ET.SubElement(activity, "PlannedFinishDate").text = task.TaskTime.ScheduleFinish
            ET.SubElement(activity, "FinishDate").text = task.TaskTime.ScheduleFinish
            self.finishes.append(task.TaskTime.ScheduleFinish)
        data_map = {
            "RemainingEarlyStartDate": task.TaskTime.EarlyStart,
            "RemainingEarlyFinishDate": task.TaskTime.EarlyFinish,
            "RemainingLateStartDate": task.TaskTime.LateStart,
            "RemainingLateFinishDate": task.TaskTime.LateFinish,
        }
        for key, value in data_map.items():
            el = ET.SubElement(activity, key)
            if value:
                el.text = value
        ET.SubElement(activity, "ProjectObjectId").text = self.id_map[work_schedule]
        ET.SubElement(activity, "ActualDuration").text = "0"
        ET.SubElement(activity, "ActualFinishDate")
        ET.SubElement(activity, "ActualStartDate")
        calendar = ifcopenshell.util.sequence.derive_calendar(task)
        if calendar:
            ET.SubElement(activity, "CalendarObjectId").text = self.id_map[calendar]
        ET.SubElement(activity, "DurationPercentComplete").text = "0"
        ET.SubElement(activity, "Type").text = "Task Dependent"
        ET.SubElement(activity, "Status").text = "Not Started"

    def create_relationships(self, tasks, work_schedule):
        for task in tasks:
            for rel in task.IsSuccessorFrom or []:
                predecessor = rel.RelatingProcess
                relationship = ET.SubElement(self.element_map[work_schedule], "Relationship")
                self.link_element(rel, relationship)
                if rel.TimeLag and rel.TimeLag.LagValue:
                    duration = ifcopenshell.util.date.ifc2datetime(rel.TimeLag.LagValue.wrappedValue)
                    ET.SubElement(relationship, "Lag").text = str(duration.days * self.hours_per_day)
                else:
                    ET.SubElement(relationship, "Lag").text = "0"
                ET.SubElement(relationship, "PredecessorProjectObjectId").text = self.id_map[work_schedule]
                ET.SubElement(relationship, "SuccessorProjectObjectId").text = self.id_map[work_schedule]
                ET.SubElement(relationship, "PredecessorActivityObjectId").text = self.id_map[predecessor]
                ET.SubElement(relationship, "SuccessorActivityObjectId").text = self.id_map[task]
                ET.SubElement(relationship, "Type").text = {
                    "START_START": "Start to Start",
                    "START_FINISH": "Start to Finish",
                    "FINISH_START": "Finish to Start",
                    "FINISH_FINISH": "Finish to Finish",
                    None: "Finish to Start",
                }[rel.SequenceType]

    def get_subtasks(self, element):
        tasks = []
        if element.is_a("IfcWorkSchedule"):
            for rel in element.Controls or []:
                tasks.extend([e for e in rel.RelatedObjects if e.is_a("IfcTask")])
        elif element.is_a("IfcTask"):
            for rel in element.IsNestedBy or []:
                tasks.extend([e for e in rel.RelatedObjects if e.is_a("IfcTask")])
        return tasks

    def link_element(self, element, el):
        self.id += 1
        ET.SubElement(el, "ObjectId").text = str(self.id)
        guid = str(uuid.UUID(ifcopenshell.guid.expand(element.GlobalId))).upper()
        ET.SubElement(el, "GUID").text = "{" + guid + "}"
        self.id_map[element] = str(self.id)
        self.element_map[element] = el
        return self.id



# Ifc4D - IFC scheduling utility
# Copyright (C) 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of Ifc4D.
#
# Ifc4D is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ifc4D is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Ifc4D.  If not, see <http://www.gnu.org/licenses/>.

import datetime
from datetime import timedelta, date
from typing import List
import ifcopenshell
import ifcopenshell.api
import ifcopenshell.util.date
import xml.etree.ElementTree as ET


class MSP2Ifc:
    def __init__(self, optionalColumns: List[str] = []):
        self.xml = None
        self.file = None
        self.ns = None
        self.work_plan = None
        self.project = {}
        self.calendars = {}
        self.tasks = {}
        self.optionalColumns = optionalColumns
        self.resources = {}
        self.RESOURCE_TYPES_MAPPING = {"1": "LABOR", "0": "MATERIAL", "2": None}

    def execute(self):
        self.parse_xml()
        self.create_ifc()

    def parse_xml(self):
        tree = ET.parse(self.xml)
        project = tree.getroot()
        self.ns = {"pr": project.tag[1:].partition("}")[0]}
        self.project["Name"] = project.findtext("pr:Name", namespaces=self.ns) or "Unnamed"
        self.project["CalendarUID"] = project.findtext("pr:CalendarUID", namespaces=self.ns) or None
        self.project["MinutesPerDay"] = project.findtext("pr:MinutesPerDay", namespaces=self.ns) or None
        self.outline_level = 0
        self.outline_parents = {}
        self.parse_task_xml(project)
        self.parse_calendar_xml(project)
        # TODO Doesn't do anything right now
        # self.parse_resources_xml(project)

    def parse_relationship_xml(self, task):
        relationships = {}
        id = 0
        if task.findall("pr:PredecessorLink", self.ns):
            for relationship in task.findall("pr:PredecessorLink", self.ns):
                relationships[id] = {
                    "PredecessorTask": relationship.find("pr:PredecessorUID", self.ns).text,
                    "Type": relationship.find("pr:Type", self.ns).text,
                }
                id += 1
        return relationships

    def parse_task_xml(self, project):
        if self.project["MinutesPerDay"]:
            hours_per_day = int(self.project["MinutesPerDay"]) / 60
        else:
            hours_per_day = 8

        for task in project.find("pr:Tasks", self.ns):
            task_id = task.find("pr:UID", self.ns).text
            task_index_level = task.find("pr:OutlineLevel", self.ns).text
            wbs_id = task.find("pr:WBS", self.ns).text
            relationships = self.parse_relationship_xml(task)
            outline_level = int(task.find("pr:OutlineLevel", self.ns).text)

            if outline_level != 0:
                parent_task = self.tasks[self.outline_parents[outline_level - 1]]
                parent_task["subtasks"].append(task_id)
            self.outline_level = outline_level
            self.outline_parents[outline_level] = task_id

            # Microsoft Project stores durations in terms of hours.
            duration = ifcopenshell.util.date.ifc2datetime(task.find("pr:Duration", self.ns).text)
            hours = duration.days * 24
            hours += duration.seconds / 60 / 60
            # Let's convert it into days, where days is the appropriate hours per day
            duration = timedelta(days=hours / float(hours_per_day))

            self.tasks[task_id] = {
                "Name": task.find("pr:Name", self.ns).text,
                "OutlineNumber": task.find("pr:OutlineNumber", self.ns).text,
                "OutlineLevel": outline_level,
                "Start": datetime.datetime.fromisoformat(task.find("pr:Start", self.ns).text),
                "Finish": datetime.datetime.fromisoformat(task.find("pr:Finish", self.ns).text),
                "Duration": duration,
                "Priority": task.find("pr:Priority", self.ns).text,
                "CalendarUID": task.find("pr:CalendarUID", self.ns).text,
                "PredecessorTasks": relationships if relationships else None,
                "subtasks": [],
                "ifc": None,
            }

            # retrieve optional columns
            # If first column = "all" then retrieve all columns
            if len(self.optionalColumns) and self.optionalColumns[0] == "all":
                self.optionalColumns = [child.tag.split("}")[1] for child in task]
            
            for column in self.optionalColumns:
                if not self.tasks[task_id].get(column):
                    self.tasks[task_id][column] = task.find(f"pr:{column}", self.ns).text if task.find(f"pr:{column}", self.ns) else None
            



    def parse_calendar_xml(self, project):
        def parse_working_times(day):
            working_times = []
            if day.find("pr:WorkingTimes", self.ns):
                for working_time in day.find("pr:WorkingTimes", self.ns).findall("pr:WorkingTime", self.ns):
                    if working_time.find("pr:FromTime", self.ns) is None:
                        continue
                    working_times.append(
                        {
                            "Start": datetime.time.fromisoformat(working_time.find("pr:FromTime", self.ns).text),
                            "Finish": datetime.time.fromisoformat(working_time.find("pr:ToTime", self.ns).text),
                        }
                    )
            return working_times

        def parse_exception(exception):
            work_times = parse_working_times(exception)
            time_period = exception.find("pr:TimePeriod", self.ns)
            data = {
                "Name": exception.find("pr:Name", self.ns).text
                if exception.find("pr:Name", self.ns) is not None
                else None,
                "FromDate": datetime.datetime.fromisoformat(time_period.find("pr:FromDate", self.ns).text)
                if time_period is not None
                else None,
                "ToDate": datetime.datetime.fromisoformat(time_period.find("pr:ToDate", self.ns).text)
                if time_period is not None
                else None,
                "Occurrences": int(exception.find("pr:Occurrences", self.ns).text)
                if exception.find("pr:Occurrences", self.ns) is not None
                else None,
                "Month": exception.find("pr:Month", self.ns).text
                if exception.find("pr:Month", self.ns) is not None
                else None,
                "MonthDay": exception.find("pr:MonthDay", self.ns).text
                if exception.find("pr:MonthDay", self.ns) is not None
                else None,
                "Type": exception.find("pr:Type", self.ns).text
                if exception.find("pr:Type", self.ns) is not None
                else None,
                "WorkingTimes": work_times,
                "ifc": None,
            }
            return data

        for calendar in project.find("pr:Calendars", self.ns).findall("pr:Calendar", self.ns):
            calendar_id = calendar.find("pr:UID", self.ns).text
            week_days = []
            exceptions = []
            week_days_element = calendar.find("pr:WeekDays", self.ns)
            week_day_elements = week_days_element.findall("pr:WeekDay", self.ns) if week_days_element else []
            for week_day in week_day_elements:
                if week_day.find("pr:WorkingTimes", self.ns):
                    if week_day.find("pr:DayType", self.ns).text == "0":
                        data = parse_exception(week_day)
                        data["Type"] = "2"
                        exceptions.append(data)
                    else:
                        week_days.append(
                            {
                                "DayType": week_day.find("pr:DayType", self.ns).text,
                                "WorkingTimes": parse_working_times(week_day),
                                "ifc": None,
                            }
                        )
            exceptions_element = calendar.find("pr:Exceptions", self.ns)
            for exception in exceptions_element.findall("pr:Exception", self.ns) if exceptions_element else []:
                data = parse_exception(exception)
                exceptions.append(data)

            self.calendars[calendar_id] = {
                "Name": calendar.find("pr:Name", self.ns).text,
                "StandardWorkWeek": week_days,
                "HolidayOrExceptions": exceptions,
            }

    def create_ifc(self):
        if not self.file:
            self.create_boilerplate_ifc()
        if not self.work_plan:
            self.work_plan = ifcopenshell.api.run("sequence.add_work_plan", self.file)
        work_schedule = self.create_work_schedule()
        self.create_calendars()
        self.create_tasks(work_schedule)
        self.create_rel_sequences()

    def create_boilerplate_ifc(self):
        self.file = ifcopenshell.file(schema="IFC4")
        ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcProject")
        self.work_plan = self.file.create_entity("IfcWorkPlan")

    def create_tasks(self, work_schedule):
        for task_id in self.tasks:
            task = self.tasks[task_id]
            # Outline Level can be None or 0
            if not task["OutlineLevel"]:
                self.create_task(task, work_schedule=work_schedule)

    def create_work_schedule(self):
        return ifcopenshell.api.run(
            "sequence.add_work_schedule", self.file, name=self.project["Name"], work_plan=self.work_plan
        )

    def create_calendars(self):
        def has_work_or_exceptions(calendar):
            return calendar["StandardWorkWeek"] or calendar["HolidayOrExceptions"]

        for calendar in self.calendars.values():
            if not has_work_or_exceptions(calendar):
                continue
            calendar["ifc"] = ifcopenshell.api.run("sequence.add_work_calendar", self.file, name=calendar["Name"])
            self.process_working_week(calendar["StandardWorkWeek"], calendar["ifc"])
            self.process_exceptions(calendar["HolidayOrExceptions"], calendar["ifc"])

    def create_task(self, task, work_schedule=None, parent_task=None):
        task["ifc"] = ifcopenshell.api.run(
            "sequence.add_task",
            self.file,
            work_schedule=work_schedule if work_schedule else None,
            parent_task=parent_task["ifc"] if parent_task else None,
        )

        calendar = None
        if task["CalendarUID"] != "-1":
            calendar = self.calendars[task["CalendarUID"]]["ifc"]
        elif not parent_task and self.project["CalendarUID"]:
            calendar = self.calendars[self.project["CalendarUID"]]["ifc"]

        if calendar:
            ifcopenshell.api.run(
                "control.assign_control",
                self.file,
                **{
                    "relating_control": calendar,
                    "related_object": task["ifc"],
                },
            )

        ifcopenshell.api.run(
            "sequence.edit_task",
            self.file,
            task=task["ifc"],
            attributes={
                "Name": task["Name"],
                "Identification": task["OutlineNumber"],
                "IsMilestone": task["Start"] == task["Finish"],
            },
        )
        task_time = ifcopenshell.api.run("sequence.add_task_time", self.file, task=task["ifc"])
        ifcopenshell.api.run(
            "sequence.edit_task_time",
            self.file,
            task_time=task_time,
            attributes={
                "ScheduleStart": task["Start"],
                "ScheduleFinish": task["Finish"],
                "DurationType": "WORKTIME" if task["Duration"] else None,
                "ScheduleDuration": task["Duration"] if task["Duration"] else None,
            },
        )
        for subtask_id in task["subtasks"]:
            self.create_task(self.tasks[subtask_id], parent_task=task)


        # create pset for optional columns
        if len(self.optionalColumns):
            pset = ifcopenshell.api.run("pset.add_pset", self.file, product=task["ifc"] , name="Pset_MSP_Task")

            ifcopenshell.api.run(
                "pset.edit_pset",
                self.file,
                pset=pset,
                properties= {name: str(task[name]) for name in self.optionalColumns if task[name]}
            )

    def process_working_week(self, week, calendar):
        day_map = {
            "1": 7,  # Sunday
            "2": 1,  # Monday
            "3": 2,  # Tuesday
            "4": 3,  # Wednesday
            "5": 4,  # Thursday
            "6": 5,  # Friday
            "7": 6,  # Saturday
            "0": 0,  # Exception
        }
        for day in week:
            if day["ifc"]:
                continue

            day["ifc"] = ifcopenshell.api.run(
                "sequence.add_work_time", self.file, work_calendar=calendar, time_type="WorkingTimes"
            )

            weekday_component = [day_map[day["DayType"]]]
            for day2 in week:
                if day["DayType"] == day2["DayType"]:
                    continue
                if day["WorkingTimes"] == day2["WorkingTimes"]:
                    weekday_component.append(day_map[day2["DayType"]])
                    # Don't process the next day, as we can group it
                    day2["ifc"] = day["ifc"]

            work_time_name = "Weekdays: {}".format(", ".join([str(c) for c in sorted(weekday_component)]))
            ifcopenshell.api.run(
                "sequence.edit_work_time",
                self.file,
                work_time=day["ifc"],
                attributes={"Name": work_time_name},
            )

            recurrence = ifcopenshell.api.run(
                "sequence.assign_recurrence_pattern", self.file, parent=day["ifc"], recurrence_type="WEEKLY"
            )
            ifcopenshell.api.run(
                "sequence.edit_recurrence_pattern",
                self.file,
                recurrence_pattern=recurrence,
                attributes={"WeekdayComponent": weekday_component},
            )
            for work_time in day["WorkingTimes"]:
                ifcopenshell.api.run(
                    "sequence.add_time_period",
                    self.file,
                    recurrence_pattern=recurrence,
                    start_time=work_time["Start"],
                    end_time=work_time["Finish"],
                )

    def create_rel_sequences(self):
        self.sequence_type_map = {
            "0": "FINISH_FINISH",
            "1": "FINISH_START",
            "2": "START_FINISH",
            "3": "START_START",
        }
        for task in self.tasks.values():
            if not task["PredecessorTasks"]:
                continue
            for predecessor in task["PredecessorTasks"].values():
                rel_sequence = ifcopenshell.api.run(
                    "sequence.assign_sequence",
                    self.file,
                    related_process=task["ifc"],
                    relating_process=self.tasks[predecessor["PredecessorTask"]]["ifc"],
                )
                if predecessor["Type"]:
                    ifcopenshell.api.run(
                        "sequence.edit_sequence",
                        self.file,
                        rel_sequence=rel_sequence,
                        attributes={"SequenceType": self.sequence_type_map[predecessor["Type"]]},
                    )

    def parse_resources_xml(self, project):
        resources_lst = project.find("pr:Resources", self.ns)
        resources = resources_lst.findall("pr:Resource", self.ns)
        for resource in resources:
            name = resource.find("pr:Name", self.ns)
            id = resource.find("pr:ID", self.ns).text
            if name is not None:
                name = name.text
            else:
                name = None
            self.resources[id] = {
                "Name": name,
                "Code": resource.find("pr:UID", self.ns).text,
                "ParentObjectId": None,
                "Type": self.RESOURCE_TYPES_MAPPING[resource.find("pr:Type", self.ns).text],
                "ifc": None,
                "rel": None,
            }

    def process_exceptions(self, exceptions, calendar):
        for exception in exceptions or []:
            self.process_exception(exception, calendar)

    def process_exception(self, exception, calendar):
        if exception["ifc"] or not exception["FromDate"]:
            return
        exception["ifc"] = ifcopenshell.api.run(
            "sequence.add_work_time", self.file, work_calendar=calendar, time_type="ExceptionTimes"
        )
        ifcopenshell.api.run(
            "sequence.edit_work_time",
            self.file,
            work_time=exception["ifc"],
            attributes={
                "Name": exception["Name"],
                "Start": ifcopenshell.util.date.datetime2ifc(exception["FromDate"], "IfcDate"),
                "Finish": ifcopenshell.util.date.datetime2ifc(exception["ToDate"], "IfcDate"),
            },
        )
        # BIG assumptions due to missing types enumeration in docs https://learn.microsoft.com/en-us/office-project/xml-data-interchange/exception-element?view=project-client-2016
        recurrence_type = None
        if exception["Type"] == "1":
            recurrence_type = "DAILY"
            attributes = {
                "Occurrences": int(exception["Occurrences"]) if exception["Occurrences"] else None,
            }
        elif exception["Type"] == "2":
            recurrence_type = "YEARLY_BY_DAY_OF_MONTH"
            month_component = [int(exception["Month"]) + 1] if exception["Month"] else None
            day_component = [int(exception["MonthDay"])] if exception["MonthDay"] else None
            if month_component is None and (exception["FromDate"].date().day == exception["ToDate"].date().day):
                month_component = [exception["FromDate"].date().month]
                day_component = [exception["FromDate"].date().day]
            attributes = {
                "MonthComponent": month_component,
                "DayComponent": day_component,
                "Occurrences": int(exception["Occurrences"]) if exception["Occurrences"] else None,
            }
        else:
            return
        recurrence = ifcopenshell.api.run(
            "sequence.assign_recurrence_pattern",
            self.file,
            parent=exception["ifc"],
            recurrence_type=recurrence_type,
        )
        ifcopenshell.api.run(
            "sequence.edit_recurrence_pattern", self.file, recurrence_pattern=recurrence, attributes=attributes
        )
        for work_time in exception["WorkingTimes"] or []:
            ifcopenshell.api.run(
                "sequence.add_time_period",
                self.file,
                recurrence_pattern=recurrence,
                start_time=work_time["Start"],
                end_time=work_time["Finish"],
            )
